import base64
import hashlib
from io import BytesIO

import pyvips
from celery import Celery
from celery.utils.log import get_task_logger
from minio import Minio
from sqlalchemy import select, update

from foxnouns.db import Member, User
from foxnouns.db.sync import session
from foxnouns.settings import MINIO, REDIS_URL

app = Celery("tasks", broker=REDIS_URL)

logger = get_task_logger(__name__)

minio = Minio(
    MINIO["ENDPOINT"],
    access_key=MINIO["ACCESS_KEY"],
    secret_key=MINIO["SECRET_KEY"],
    secure=MINIO["SECURE"],
    region=MINIO["REGION"],
)
bucket = MINIO["BUCKET"]


def convert_avatar(uri: str) -> bytes:
    """Converts a base64 data URI into a WebP image.
    Images are resized and cropped to 512x512 and exported with quality 95.
    Only PNG, WebP, and JPEG images are allowed as input."""

    if not uri.startswith("data:image/"):
        raise ValueError("Not a data URI")

    content_type, encoded = uri.removeprefix("data:").split(";base64,", 1)
    if content_type not in ["image/png", "image/webp", "image/jpeg"]:
        raise ValueError("Invalid content type for image")

    img = pyvips.Image.thumbnail_buffer(
        base64.b64decode(encoded),
        512,
        height=512,
        size=pyvips.Size.BOTH,
        crop=pyvips.Interesting.CENTRE,
    )
    return img.write_to_buffer(".webp", Q=95)


@app.task
def process_user_avatar(user_id: int, avatar: str) -> None:
    """Processes an avatar string, uploads it to S3, and updates the user's avatar hash.
    Also deletes the old avatar if one was already set."""

    with session() as conn:
        user = conn.scalar(select(User).where(User.id == user_id))
    if not user:
        raise ValueError("process_user_avatar was passed the ID of a nonexistent user")

    img = convert_avatar(avatar)
    hash = hashlib.new("sha256", data=img).hexdigest()
    old_hash = user.avatar

    minio.put_object(
        bucket,
        f"users/{user_id}/avatars/{hash}.webp",
        BytesIO(img),
        len(img),
        "image/webp",
    )

    with session() as conn:
        conn.execute(update(User).values(avatar=hash).where(User.id == user_id))
        conn.commit()

    if old_hash and old_hash != hash:
        minio.remove_object(bucket, f"users/{user_id}/avatars/{old_hash}.webp")


@app.task
def delete_user_avatar(user_id: int) -> None:
    """Deletes a user's avatar."""

    with session() as conn:
        user = conn.scalar(select(User).where(User.id == user_id))
    if not user:
        raise ValueError("delete_user_avatar was passed the ID of a nonexistent user")
    if not user.avatar:
        logger.info(
            "delete_user_avatar was called for a user with a null avatar (%d)", user_id
        )
        return

    minio.remove_object(bucket, f"users/{user_id}/avatars/{user.avatar}.webp")

    with session() as conn:
        conn.execute(update(User).values(avatar=None).where(User.id == user_id))
        conn.commit()


@app.task
def process_member_avatar(member_id: int, avatar: str) -> None:
    """Processes an avatar string, uploads it to S3, and updates the member's avatar hash.
    Also deletes the old avatar if one was already set."""

    with session() as conn:
        member = conn.scalar(select(Member).where(Member.id == member_id))
    if not member:
        raise ValueError(
            "process_member_avatar was passed the ID of a nonexistent member"
        )

    img = convert_avatar(avatar)
    hash = hashlib.new("sha256", data=img).hexdigest()
    old_hash = member.avatar

    minio.put_object(
        bucket,
        f"members/{member_id}/avatars/{hash}.webp",
        BytesIO(img),
        len(img),
        "image/webp",
    )

    with session() as conn:
        conn.execute(update(Member).values(avatar=hash).where(Member.id == member_id))
        conn.commit()

    if old_hash and old_hash != hash:
        minio.remove_object(bucket, f"members/{member_id}/avatars/{old_hash}.webp")


@app.task
def delete_member_avatar(member_id: int) -> None:
    """Deletes a member's avatar."""

    with session() as conn:
        member = conn.scalar(select(Member).where(Member.id == member_id))
    if not member:
        raise ValueError(
            "delete_member_avatar was passed the ID of a nonexistent member"
        )
    if not member.avatar:
        logger.info(
            "delete_member_avatar was called for a member with a null avatar (%d)",
            member_id,
        )
        return

    minio.remove_object(bucket, f"members/{member_id}/avatars/{member.avatar}.webp")

    with session() as conn:
        conn.execute(update(Member).values(avatar=None).where(Member.id == member_id))
        conn.commit()
